using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using System;
using System.Collections.Generic;

namespace MacGruber
{
	public partial class IdlePoser : MVRScript
	{
		private List<JSONStorableParam> myMenuStorables = new List<JSONStorableParam>();
		private List<UIDynamic> myMenuElements = new List<UIDynamic>();

		private JSONStorableUrl myDataFile;
		private JSONStorableStringChooser myMainMenu;
		private JSONStorableStringChooser myMainState;
		private JSONStorableStringChooser myCapturesControlList;
		private JSONStorableBool myStateAutoTransition;
		private JSONStorableStringChooser myAnchorCaptureList;
		private JSONStorableStringChooser myAnchorModeList;
		private JSONStorableStringChooser myAnchorAtomListA;
		private JSONStorableStringChooser myAnchorAtomListB;
		private JSONStorableStringChooser myAnchorControlListA;
		private JSONStorableStringChooser myAnchorControlListB;
		private JSONStorableFloat myAnchorBlendRatio;
		private JSONStorableFloat myAnchorDampingTime;
		private JSONStorableStringChooser myTransitionList;

		private List<string> myAnchorModes;

		private const string FILE_EXTENSION = "idlepose";


		// =======================================================================================


		private void InitUI()
		{
			myDataFile = new JSONStorableUrl("IdlePose", "", UILoadJSON, FILE_EXTENSION, true);
			RegisterUrl(myDataFile);


			List<string> menuItems = new List<string>();
			menuItems.Add("Play");
			menuItems.Add("Captures");
			menuItems.Add("State");
			menuItems.Add("Anchors");
			menuItems.Add("Transitions");
			myMainMenu = new JSONStorableStringChooser("Menu", menuItems, menuItems[0], "Menu");
			myMainMenu.setCallbackFunction += UISelectMenu;

			List<string> stateItems = new List<string>();
			myMainState = new JSONStorableStringChooser("State", stateItems, "", "State");
			myMainState.setCallbackFunction += UISelectStateAndRefresh;

			myCapturesControlList = new JSONStorableStringChooser("Control", new List<string>(), "", "Control");
			myStateAutoTransition = new JSONStorableBool("Auto-Transition on State Change", true);
			myTransitionList = new JSONStorableStringChooser("Transition", new List<string>(), "", "Transition");
			myAnchorCaptureList = new JSONStorableStringChooser("ControlCapture", new List<string>(), "", "Control Capture");
			myAnchorCaptureList.setCallbackFunction += (string v) => UIRefreshMenu();

			myAnchorModes = new List<string>();
			myAnchorModes.Add("World");
			myAnchorModes.Add("Single Anchor");
			myAnchorModes.Add("Blend Anchor");
			myAnchorModeList = new JSONStorableStringChooser("AnchorMode", myAnchorModes, "", "Anchor Mode");
			myAnchorModeList.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorAtomListA = new JSONStorableStringChooser("AnchorAtomA", new List<string>(), "", "Anchor Atom A");
			myAnchorAtomListA.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorAtomListB = new JSONStorableStringChooser("AnchorAtomB", new List<string>(), "", "Anchor Atom B");
			myAnchorAtomListB.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorControlListA = new JSONStorableStringChooser("AnchorControlA", new List<string>(), "", "Anchor Control A");
			myAnchorControlListA.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorControlListB = new JSONStorableStringChooser("AnchorControlB", new List<string>(), "", "Anchor Control B");
			myAnchorControlListB.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorBlendRatio = new JSONStorableFloat("Anchor Blend Ratio", 0.5f, 0.0f, 1.0f, true, true);
			myAnchorBlendRatio.setCallbackFunction += (float v) => UISetAnchors();
			myAnchorDampingTime = new JSONStorableFloat("Anchor Damping Time", 0.2f, 0.0f, 5.0f, true, true);
			myAnchorDampingTime.setCallbackFunction += (float v) => UISetAnchors();


			CreateMainUI();
			UIRefreshMenu();
		}

		private void UIRefreshMenu()
		{
			myMainMenu.setCallbackFunction(myMainMenu.val);
		}

		private void UISelectMenu(string menuItem)
		{
			myPaused = true;
			myMainState.popup.visible = true; // Workaround for PopupPanel appearing behind other UI for some reason
			myMainState.popup.visible = false;
			CleanupMenu();

			List<string> states = new List<string>();
			foreach (var state in myStates)
				states.Add(state.Value.myName);
			states.Sort();
			myMainState.choices = states;
			if (states.Count == 0)
			{
				myMainState.valNoCallback = "";
			}
			else if (!states.Contains(myMainState.val))
			{
				myMainState.valNoCallback = states[0];
				UIBlendToState();
			}

			if (menuItem == "Play")
				CreatePlayMenu();
			else if (menuItem == "Captures")
				CreateCapturesMenu();
			else if (menuItem == "State")
				CreateStateMenu();
			else if (menuItem == "Anchors")
				CreateAnchorsMenu();
			else if (menuItem == "Transitions")
				CreateTransitionsMenu();
		}

		private void UIBlendToState()
		{
			State state;
			if (myStateAutoTransition.val && myStates.TryGetValue(myMainState.val, out state))
				BlendTransitionToState(state);
		}

		private void UISelectStateAndRefresh(string name)
		{
			UIBlendToState();
			UIRefreshMenu();
		}

		private void CreateMainUI()
		{
			CreateMenuInfo(
				"<color=#606060><size=40><b>IdlePoser</b></size>\nCreates character idle animations. " +
				"Use the <i>Menu</i> dropdown to switch between the different UI pages.</color>",
				165, true
			);

			{
				UIDynamicButton button = CreateButton("Load", false);
				UIDynamicTextField textfield = CreateTextField(myDataFile, false);
				textfield.UItext.alignment = TextAnchor.MiddleRight;
				textfield.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
				textfield.UItext.verticalOverflow = VerticalWrapMode.Truncate;
				LayoutElement layout = textfield.GetComponent<LayoutElement>();
				layout.preferredHeight = layout.minHeight = 35;
				textfield.height = 35;
				myDataFile.setCallbackFunction -= UILoadJSON;
				myDataFile.SetFilePath("Custom/Assets/");
				myDataFile.RegisterFileBrowseButton(button.button);
				myDataFile.setCallbackFunction += UILoadJSON;

				CreateMenuButton("Save As", UISaveJSONDialog, false);
			}

			CreateMenuPopup(myMainState, false);

			CreateMenuPopup(myMainMenu, true);

			CreateMenuSyncSpacer(15);

			myMenuStorables.Clear();
			myMenuElements.Clear();
		}

		private void CreatePlayMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Play Idle</b></size>", false);

			if (myStates.Count >= 1)
				CreateMenuInfo("IdlePoser is playing animations.", 100, false);
			else
				CreateMenuInfo("You need to add some states and transitions before you can play animations.", 100, false);

			myPaused = false;
		}

		private void CreateCapturesMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Control Captures</b></size>", false);
			for (int i=0; i<myControlCaptures.Count; ++i)
			{
				ControlCapture cc = myControlCaptures[i];
				CreateMenuButton(cc.myName, () => UIRemoveControlCapture(cc), false);
			}

			CreateMenuSpacer(15, false);

			FreeControllerV3[] atomControls = GetAllControllers();
			List<string> availableControls = new List<string>();
			for (int i=0; i<atomControls.Length; ++i)
			{
				string control = atomControls[i].name;
				if (control.StartsWith("hair"))
					continue;
				if (myControlCaptures.FindIndex(cc => cc.myName == control) >= 0)
					continue;
				availableControls.Add(control);
			}
			myCapturesControlList.choices = availableControls;
			if (availableControls.Count == 0)
				myCapturesControlList.val = "";
			else if (!availableControls.Contains(myCapturesControlList.val))
				myCapturesControlList.val = availableControls[0];
			CreateMenuPopup(myCapturesControlList, false);
			CreateMenuButton("Add Controller", UIAddControlCapture, false);


			CreateMenuInfoOneLine("<size=30><b>Morph Captures</b></size>", true);
		}

		private void CreateStateMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>State Manager</b></size>", false);
			CreateMenuButton("Add State", UIAddState, false);
			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
				return;

			CreateMenuButton("Remove State", UIRemoveState, false);
			CreateMenuToggle(myStateAutoTransition, false);

			CreateMenuSpacer(15, false);

			CreateMenuInfoOneLine("<size=30><b>State Settings</b></size>", false);

			JSONStorableString name = new JSONStorableString("State Name", state.myName);
			name.setCallbackFunction = UIRenameState;
			CreateMenuTextInput(name, false);

			CreateMenuButton("Capture State", UICaptureState, false);

			JSONStorableFloat transitionDuration = new JSONStorableFloat("Transition Duration", state.myTransitionDuration, 0.01f, 5.0f, true, true);
			transitionDuration.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myTransitionDuration = v;
			};
			CreateMenuSlider(transitionDuration, false);

			JSONStorableBool isTransition = new JSONStorableBool("Is Transit State", state.myIsTransition);
			isTransition.setCallbackFunction = (bool v) => {
				State s = UIGetState();
				if (s != null)
					s.myIsTransition = v;
			};
			CreateMenuToggle(isTransition, false);

			JSONStorableFloat waitDurationMin = new JSONStorableFloat("Wait Duration Min", state.myWaitDurationMin, 0.0f, 300.0f, true, true);
			waitDurationMin.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myWaitDurationMin = v;
			};
			CreateMenuSlider(waitDurationMin, false);

			JSONStorableFloat waitDurationMax = new JSONStorableFloat("Wait Duration Max", state.myWaitDurationMax, 0.0f, 300.0f, true, true);
			waitDurationMax.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myWaitDurationMax = v;
			};
			CreateMenuSlider(waitDurationMax, false);
		}

		private void CreateAnchorsMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Anchors</b></size>", false);

			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				CreateMenuInfo("You need to add a state before you can configure anchors.", 100, false);
				return;
			}

			List<string> captures = new List<string>(myControlCaptures.Count);
			for (int i=0; i<myControlCaptures.Count; ++i)
				captures.Add(myControlCaptures[i].myName);
			myAnchorCaptureList.choices = captures;
			if (captures.Count == 0)
				myAnchorCaptureList.valNoCallback = "";
			else if (!captures.Contains(myAnchorCaptureList.val))
				myAnchorCaptureList.valNoCallback = captures[0];
			CreateMenuPopup(myAnchorCaptureList, false);

			ControlCapture controlCapture = myControlCaptures.Find(cc => cc.myName == myAnchorCaptureList.val);
			if (controlCapture == null)
				return;

			CreateMenuSpacer(30, false);

			ControlEntryAnchored controlEntry;
			if (!state.myControlEntries.TryGetValue(controlCapture, out controlEntry))
				return;
			myAnchorModeList.valNoCallback = myAnchorModes[controlEntry.myAnchorMode];
			CreateMenuPopup(myAnchorModeList, false);

			if (controlEntry.myAnchorMode != ControlEntryAnchored.ANCHORMODE_WORLD)
			{
				bool isBlendMode = controlEntry.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND;

				List<string> visibleAtoms = GetAtomUIDs();
				List<string> atoms = new List<string>(visibleAtoms.Count+1);
				atoms.Add("[Self]");
				atoms.AddRange(visibleAtoms);
				myAnchorAtomListA.label = isBlendMode ? "Anchor Atom A" : "Anchor Atom";
				myAnchorAtomListA.valNoCallback = controlEntry.myAnchorAAtom;
				myAnchorAtomListA.choices = atoms;
				CreateMenuPopup(myAnchorAtomListA, false);

				Atom atomA = GetSelfOrAtomById(myAnchorAtomListA.val);
				List<string> controls = atomA?.GetStorableIDs() ?? new List<string>();
				controls.RemoveAll((string control) => {
					if (atomA == containingAtom && control == myAnchorCaptureList.val)
						return true;
					JSONStorable storable = atomA.GetStorableByID(control);
					if (storable is DAZBone || storable is FreeControllerV3)
						return false;
					return true;
				});
				myAnchorControlListA.label = isBlendMode ? "Anchor Control A" : "Anchor Control";
				myAnchorControlListA.valNoCallback = controlEntry.myAnchorAControl;
				myAnchorControlListA.choices = controls;
				if (controls.Count == 0)
					myAnchorControlListA.valNoCallback = "";
				else if (!controls.Contains(myAnchorControlListA.val))
					myAnchorControlListA.valNoCallback = controls[0];
				CreateMenuPopup(myAnchorControlListA, false);

				if (isBlendMode)
				{
					myAnchorAtomListB.valNoCallback = controlEntry.myAnchorBAtom;
					myAnchorAtomListB.choices = atoms;
					CreateMenuPopup(myAnchorAtomListB, false);

					Atom atomB = GetSelfOrAtomById(myAnchorAtomListB.val);
					controls = atomB?.GetStorableIDs() ?? new List<string>();
					controls.RemoveAll((string control) => {
						if (atomB == containingAtom && control == myAnchorCaptureList.val)
							return true;
						JSONStorable storable = atomB.GetStorableByID(control);
						if (storable is DAZBone || storable is FreeControllerV3)
							return false;
						return true;
					});
					myAnchorControlListB.valNoCallback = controlEntry.myAnchorBControl;
					myAnchorControlListB.choices = controls;
					if (controls.Count == 0)
						myAnchorControlListB.valNoCallback = "";
					else if (!controls.Contains(myAnchorControlListB.val))
						myAnchorControlListB.valNoCallback = controls[0];
					CreateMenuPopup(myAnchorControlListB, false);

					myAnchorBlendRatio.valNoCallback = controlEntry.myBlendRatio;
					CreateMenuSlider(myAnchorBlendRatio, false);
				}

				myAnchorDampingTime.valNoCallback = controlEntry.myDampingTime;
				CreateMenuSlider(myAnchorDampingTime, false);
			}
		}

		private void CreateTransitionsMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Transitions</b></size>", false);

			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				CreateMenuInfo("You need to add some states before you can add transitions.", 100, false);
				return;
			}

			for (int i=0; i<state.myTransitions.Count; ++i)
			{
				State target = state.myTransitions[i];
				CreateMenuButton(target.myName, () => UIRemoveTransition(state, target), false);
			}
			if (state.myTransitions.Count == 0)
			{
				CreateMenuInfoOneLine("There are no transitions for this state.", false);
			}

			CreateMenuSpacer(15, false);

			List<string> availableTargets = new List<string>();
			foreach (var t in myStates)
			{
				State target = t.Value;
				if (state == target)
					continue;
				if (state.myTransitions.Contains(target))
					continue;
				availableTargets.Add(target.myName);
			}
			myTransitionList.choices = availableTargets;
			if (availableTargets.Count == 0)
				myTransitionList.val = "";
			else if (!availableTargets.Contains(myTransitionList.val))
				myTransitionList.val = availableTargets[0];
			CreateMenuPopup(myTransitionList, false);
			CreateMenuButton("Add Transition (one-way)", () => UIAddTransition(true), false);
			CreateMenuButton("Add Transition (2-way)", () => UIAddTransition(false), false);
		}


		// =======================================================================================


		private void UILoadJSON(string url)
		{
			LoadJSON(url);
			if (myCurrentState != null)
			{
				myMainState.valNoCallback = myCurrentState.myName;
				myMainState.setCallbackFunction(myCurrentState.myName);
			}
			else
			{
				UIRefreshMenu();
			}
		}

		private void UISaveJSONDialog()
		{
			SuperController sc = SuperController.singleton;
			sc.GetMediaPathDialog(UISaveJSON, FILE_EXTENSION, "Custom/Assets", false, true, false, null, false, null, false, false);
			sc.mediaFileBrowserUI.SetTextEntry(true);
			if (sc.mediaFileBrowserUI.fileEntryField != null)
			{
				string filename = ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString();
				sc.mediaFileBrowserUI.fileEntryField.text = filename + "." + FILE_EXTENSION;
				sc.mediaFileBrowserUI.ActivateFileNameField();
			}
		}

		private void UISaveJSON(string path)
		{
			path = path.Replace('\\', '/');
			SuperController.LogMessage("IdlePoser: Saving as '"+path+"'.");
			SaveJSON(path);
		}

		private void UIAddControlCapture()
		{
			ControlCapture cc = new ControlCapture(this, myCapturesControlList.val);
			myControlCaptures.Add(cc);
			foreach (var state in myStates)
				cc.CaptureEntry(state.Value);
			UIRefreshMenu();
		}

		private void UIRemoveControlCapture(ControlCapture cc)
		{
			myControlCaptures.Remove(cc);
			foreach (var state in myStates)
				state.Value.myControlEntries.Remove(cc);
			UIRefreshMenu();
		}

		private void UIAddState()
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "State#" + i;
				if (!myStates.ContainsKey(name))
				{
					CreateState(name);
					myMainState.val = name;
					UIRefreshMenu();
					return;
				}
			}
			SuperController.LogError("IdlePoser: Too many states!");
		}

		private void UICaptureState()
		{
			State state = UIGetState();
			if (state == null)
				return;

			CaptureState(state);
			UIRefreshMenu();
		}

		private void UIRemoveState()
		{
			State state = UIGetState();
			if (state == null)
				return;

			myStates.Remove(myMainState.val);
			foreach (var s in myStates)
				s.Value.myTransitions.RemoveAll(x => x == s.Value);

			UIRefreshMenu();
		}

		private void UIRenameState(string name)
		{
			State state = UIGetState();
			if (state == null)
				return;
			if (state.myName == name)
				return;

			myStates.Remove(state.myName);

			int altIndex = 2;
			string baseName = name;
			while (myStates.ContainsKey(name))
			{
				name = baseName + "#" + altIndex;
				++altIndex;
			}

			myStates.Add(name, state);
			state.myName = name;
			myMainState.valNoCallback = name;
			UIRefreshMenu();
		}

		private void UISetAnchors()
		{
			State state = UIGetState();
			if (state == null)
				return;

			ControlCapture controlCapture = myControlCaptures.Find(cc => cc.myName == myAnchorCaptureList.val);
			if (controlCapture == null)
				return;

			ControlEntryAnchored controlEntry;
			if (!state.myControlEntries.TryGetValue(controlCapture, out controlEntry))
				return;

			int anchorMode = myAnchorModes.FindIndex(m => m == myAnchorModeList.val);
			if (anchorMode >= 0)
				controlEntry.myAnchorMode = anchorMode;
			controlEntry.myBlendRatio = myAnchorBlendRatio.val;
			controlEntry.myDampingTime = myAnchorDampingTime.val;

			if (anchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
			{
				controlEntry.myAnchorAAtom = myAnchorAtomListA.val;
				controlEntry.myAnchorAControl = myAnchorControlListA.val;
			}
			if (anchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
			{
				controlEntry.myAnchorBAtom = myAnchorAtomListB.val;
				controlEntry.myAnchorBControl = myAnchorControlListB.val;
			}
			controlEntry.AdjustAnchor(this);

			UIRefreshMenu();
		}

		private void UIAddTransition(bool oneWay)
		{
			State source = UIGetState();
			if (source == null)
				return;
			State target;
			if (!myStates.TryGetValue(myTransitionList.val, out target))
				return;

			if (!source.myTransitions.Contains(target))
				source.myTransitions.Add(target);
			if (!oneWay && !target.myTransitions.Contains(source))
				target.myTransitions.Add(source);

			UIRefreshMenu();
		}

		private void UIRemoveTransition(State source, State target)
		{
			source.myTransitions.Remove(target);
			UIRefreshMenu();
		}

		private State UIGetState()
		{
			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				SuperController.LogError("IdlePoser: Invalid state selected!");
				return null;
			}
			else
			{
				return state;
			}
		}

		// =======================================================================================


		private void CreateMenuButton(string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
			myMenuElements.Add(button);
		}

		private void CreateMenuToggle(JSONStorableBool storable, bool rightSide)
		{
			CreateToggle(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuSlider(JSONStorableFloat storable, bool rightSide)
		{
			CreateSlider(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuPopup(JSONStorableStringChooser storable, bool rightSide)
		{
			CreateScrollablePopup(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuTextInput(JSONStorableString storable, bool rightSide)
        {
            UIDynamicTextField textfield = CreateTextField(storable, rightSide);
            textfield.height = 35;
            textfield.backgroundColor = Color.white;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
            InputField inputfield = textfield.gameObject.AddComponent<InputField>();
            inputfield.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
            inputfield.textComponent = textfield.UItext;
            storable.inputField = inputfield;
			myMenuStorables.Add(storable);
        }

		private void CreateMenuInfo(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			myMenuStorables.Add(storable);
		}

		private void CreateMenuInfoOneLine(string text, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.UItext.alignment = TextAnchor.LowerLeft;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
			myMenuStorables.Add(storable);
		}

		private void CreateMenuSpacer(float height, bool rightSide)
		{
			UIDynamic spacer = CreateSpacer(rightSide);
			spacer.height = height;
			myMenuElements.Add(spacer);
		}

		private void CreateMenuSyncSpacer(float height)
		{
			float SPACING = 15;
			float leftHeight = 0.0f;
			if (leftUIElements.Count >= 1)
				leftHeight += (leftUIElements.Count-1) * SPACING;
			for (int i=0; i < leftUIElements.Count; ++i)
			{
				RectTransform rt = leftUIElements[i] as RectTransform;
				if (rt != null)
					leftHeight += rt.sizeDelta.y;
			}
			float rightHeight = 0.0f;
			if (rightUIElements.Count >= 1)
				rightHeight += (rightUIElements.Count-1) * SPACING;
			for (int i=0; i < rightUIElements.Count; ++i)
			{
				RectTransform rt = rightUIElements[i] as RectTransform;
				if (rt != null)
					rightHeight += rt.sizeDelta.y;
			}

			height = Mathf.Max(height, 0.0f);
			bool extraHeight = height > 0;
			bool side = leftHeight > rightHeight;
			UIDynamic spacer = CreateSpacer(side);
			spacer.height = Mathf.Abs(leftHeight - rightHeight) + (extraHeight ? height : -SPACING);
			myMenuElements.Add(spacer);
			if (extraHeight)
			{
				spacer = CreateSpacer(!side);
				spacer.height = height;
				myMenuElements.Add(spacer);
			}
		}

		private void CleanupMenu()
		{
			for (int i=0; i<myMenuStorables.Count; ++i)
			{
				JSONStorableParam jsp = myMenuStorables[i];
				if (jsp is JSONStorableFloat)
					RemoveSlider(jsp as JSONStorableFloat);
				else if (jsp is JSONStorableBool)
					RemoveToggle(jsp as JSONStorableBool);
				else if (jsp is JSONStorableColor)
					 RemoveColorPicker(jsp as JSONStorableColor);
				else if (jsp is JSONStorableString)
					RemoveTextField(jsp as JSONStorableString);
				else if (jsp is JSONStorableStringChooser)
				{
					// Workaround for VaM not cleaning its panels properly.
					JSONStorableStringChooser jssc = jsp as JSONStorableStringChooser;
					RectTransform popupPanel = jssc.popup?.popupPanel;
					RemovePopup(jssc);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
			}
			myMenuStorables.Clear();

			for (int i=0; i<myMenuElements.Count; ++i)
			{
				UIDynamic uid = myMenuElements[i];
				if (uid is UIDynamicButton)
					RemoveButton(uid as UIDynamicButton);
				else if (uid is UIDynamicSlider)
					RemoveSlider(uid as UIDynamicSlider);
				else if (uid is UIDynamicToggle)
					RemoveToggle(uid as UIDynamicToggle);
				else if (uid is UIDynamicColorPicker)
					 RemoveColorPicker(uid as UIDynamicColorPicker);
				else if (uid is UIDynamicTextField)
					RemoveTextField(uid as UIDynamicTextField);
				else if (uid is UIDynamicPopup)
				{
					// Workaround for VaM not cleaning its panels properly.
					UIDynamicPopup uidp = uid as UIDynamicPopup;
					RectTransform popupPanel = uidp.popup?.popupPanel;
					RemovePopup(uidp);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
				else
					RemoveSpacer(uid);
			}
			myMenuElements.Clear();
		}
	}

}